/*
 * iSCSI login library
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-login.c,v 1.19 2002/10/08 20:32:33 smferris Exp $
 *
 */

#include "iscsi-platform.h"
#include "iscsi-protocol.h"
#include "iscsi-io.h"
#include "iscsi-login.h"

struct IscsiHdr *iscsi_align_pdu(iscsi_session_t *session, unsigned char *buffer, int buffersize)
{
    struct IscsiHdr *header;
    unsigned long addr = (unsigned long)buffer;

    /* find a buffer location guaranteed to be reasonably aligned for the header */
    addr += (addr % sizeof(*header));
    header = (struct IscsiHdr *)addr;

    return header;
}


/* caller is assumed to be well-behaved and passing NUL terminated strings */
int iscsi_add_text(iscsi_session_t *session, struct IscsiHdr *pdu, char *data, int max_data_length, char *param, char *value)
{
    int param_len = strlen(param);
    int value_len = strlen(value);
    int length = param_len + 1 + value_len + 1; /* param, separator, value, and trailing NUL */
    int pdu_length = ntoh24(pdu->dlength);
    char *text = data;
    char *end = data + max_data_length;
    char *pdu_text;
    
    /* find the end of the current text */
    text += pdu_length;
    pdu_text = text;
    pdu_length += length;

    if (text + length >= end) {
        logmsg(AS_NOTICE, "failed to add login text '%s=%s'\n", param, value);
        return 0;
    }

    /* param */
    iscsi_strncpy(text, param, param_len);
    text += param_len;

    /* separator */
    *text++ = ISCSI_TEXT_SEPARATOR;

    /* value */
    strncpy(text, value, value_len);
    text += value_len;
    
    /* NUL */
    *text++ = '\0';

    /* update the length in the PDU header */
    hton24(pdu->dlength, pdu_length);

    return 1;
}

int iscsi_find_key_value(char *param, char *pdu, char *pdu_end, char **value_start, char **value_end)
{
    char *str = param;
    char *text = pdu;
    char *value = NULL;

    if (value_start)
        *value_start = NULL;
    if (value_end)
        *value_end = NULL;
        
    /* make sure they contain the same bytes */
    while (*str) {
        if (text >= pdu_end) 
            return 0;
        if (*text == '\0')
            return 0;
        if (*str != *text)
            return 0;
        str++;
        text++;
    }
    
    if ((text >= pdu_end) || (*text == '\0') || (*text != ISCSI_TEXT_SEPARATOR)) {
        return 0;
    }

    /* find the value */
    value = text + 1;

    /* find the end of the value */
    while ((text < pdu_end) && (*text))
        text++;

    if (value_start)
        *value_start = value;
    if (value_end)
        *value_end = text;

    return 1;
}

/* we never authenticate the target, so we don't need an AuthClient callback that does anything.
 * It would only be needed if we were authenticating the target, which we currently never do.
 */
static void null_callback(void *user_handle, void *message_handle, int auth_status)
{
}


/* this assumes the text data is always NUL terminated.  The caller can always arrange for that
 * by using a slightly larger buffer than the max PDU size, and then appending a NUL to the PDU.
 */
int iscsi_process_login_response(iscsi_session_t *session, struct IscsiLoginRspHdr *login_rsp_pdu, char *data, int max_data_length)
{
    int tbit = login_rsp_pdu->tbit;
    char *text = data;
    char *end;

    end = text + ntoh24(login_rsp_pdu->dlength) + 1; 
    if (end >= (data + max_data_length)) {
        logmsg(AS_ERROR, "process_login_response - buffer too small to guarantee NUL termination\n");
        return 0;
    }
    /* guarantee a trailing NUL */
    *end = '\0';

    /* if the response status was success, sanity check the response */
    if (login_rsp_pdu->status_class == STATUS_CLASS_SUCCESS) {
        /* check the active version */
        if (login_rsp_pdu->active_version != ISCSI_DRAFT16_VERSION) {
            logmsg(AS_ERROR, "received incompatible active iSCSI version 0x%02x during login, expected version 0x%02x\n",
                   login_rsp_pdu->active_version, ISCSI_DRAFT16_VERSION);
            return 0;
        }

        /* make sure the current phase matches */
        if (login_rsp_pdu->curr != session->current_phase) {
            logmsg(AS_ERROR, "current phase mismatch, session %d, response %d\n", 
                   session->current_phase, login_rsp_pdu->curr);
            return 0;
        }

        /* make sure that we're actually advancing if the T-bit is set */
        if (login_rsp_pdu->tbit && (login_rsp_pdu->next <= session->current_phase)) {
            logmsg(AS_ERROR, "current phase %d, target wants to go to phase %d, but we want to go to phase %d\n",
                   session->current_phase, login_rsp_pdu->next, session->next_phase);
            return 0;
        }
    }

    if (session->current_phase == ISCSI_SECURITY_NEGOTIATION_PHASE) {
        if (iscsiAuthClientRecvBegin(session->auth_client) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "authClientRecvBegin failed\n");
            return 0;
        }
        
        if (iscsiAuthClientRecvTransitBit(session->auth_client, tbit) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "authClientRecvTransitBit failed\n");
            return 0;
        }
    }

    /* scan the text data */
 more_text:
    while (text && (text < end)) {
        char *value = NULL;
        char *value_end = NULL;

        /* skip any NULs separating each text key=value pair */
        while ((text < end) && (*text == '\0'))
            text++;
        if (text >= end)
            break;

        /* handle keys appropriate for each phase */
        switch (session->current_phase) {
            case ISCSI_SECURITY_NEGOTIATION_PHASE: {
                /* a few keys are possible in Security phase which the
                 * auth code doesn't care about, but which we might
                 * want to see, or at least not choke on.
                 */
                if (iscsi_find_key_value("TargetAlias", text, end, &value, &value_end)) {
                    size_t size = sizeof(session->TargetAlias);

                    if ((value_end - value) < size)
                        size = value_end - value;

                    memcpy(session->TargetAlias, value, size);
                    session->TargetAlias[sizeof(session->TargetAlias)-1] = '\0';
                    text = value_end;
                }
                else if (iscsi_find_key_value("TargetAddress", text, end, &value, &value_end)) {
                    /* if possible, change the session's ip_address and port to the new TargetAddress */
                    if (session->update_address && session->update_address(session, value)) {
                        text = value_end;
                    }
                    else {
                        logmsg(AS_ERROR, "session %p can't handle redirection to %s\n", session, value);
                        return 0;
                    }
                }
                else if (iscsi_find_key_value("TargetPortalGroupTag", text, end, &value, &value_end)) {
                    /* We should have already obtained this via discovery. 
                     * We've already picked an isid, so the most we can do is confirm we reached
                     * the portal group we were expecting to.
                     */
                    int tag = iscsi_strtoul(value, NULL, 0);
                    if (session->portal_group_tag >= 0) {
                        if (tag != session->portal_group_tag) {
                            logmsg(AS_ERROR, "session %p portal group tag mismatch, expected %u, actual %u\n",
                                   session, session->portal_group_tag, tag);
                            return 0;
                        }
                    }
                    else {
                        /* we now know the tag */
                        session->portal_group_tag = tag;
                    }
                    text = value_end;
                } 
                else {
                    /* any key we don't recognize either goes to the auth code, or we choke on it */
                    int keytype = iscsiAuthKeyTypeNone;
                    
                    while (iscsiAuthClientGetNextKeyType(&keytype) == iscsiAuthStatusNoError) {
                        char *key = (char *)iscsiAuthClientGetKeyName(keytype);
                        
                        if (key && iscsi_find_key_value(key, text, end, &value, &value_end)) {
                            if (iscsiAuthClientRecvKeyValue(session->auth_client, keytype, value) != iscsiAuthStatusNoError) {
                                logmsg(AS_ERROR, "authentication failed to process %s\n", text);
                                return 0;
                            }
                            text = value_end;
                            goto more_text;
                        }
                    }

                    logmsg(AS_ERROR, "security phase failed to recognize text %s\n", text);
                    return 0;
                }
                break;
            }
            case ISCSI_OP_PARMS_NEGOTIATION_PHASE: {
                /* FIXME: they're making base64 an encoding option for
                 * all numbers in draft13, since some security
                 * protocols use large numbers, and it was somehow
                 * considered "simpler" to let them be used for any
                 * number anywhere.
                 */
                if (iscsi_find_key_value("TargetAlias", text, end, &value, &value_end)) {
                    size_t size = sizeof(session->TargetAlias);

                    if ((value_end - value) < size)
                        size = value_end - value;

                    memcpy(session->TargetAlias, value, size);
                    session->TargetAlias[sizeof(session->TargetAlias)-1] = '\0';
                    text = value_end;
                }
                else if (iscsi_find_key_value("TargetAddress", text, end, &value, &value_end)) {
                    if (session->update_address && session->update_address(session, value)) {
                        text = value_end;
                    }
                    else {
                        logmsg(AS_ERROR, "session %p can't handle redirection to %s\n", session, value);
                        return 0;
                    }
                }
                else if (iscsi_find_key_value("TargetPortalGroupTag", text, end, &value, &value_end)) {
                    /* We should have already obtained this via discovery. 
                     * We've already picked an isid, so the most we can do is confirm we reached
                     * the portal group we were expecting to.
                     */
                    int tag = iscsi_strtoul(value, NULL, 0);
                    if (session->portal_group_tag >= 0) {
                        if (tag != session->portal_group_tag) {
                            logmsg(AS_ERROR, "session %p portal group tag mismatch, expected %u, actual %u\n",
                                   session, session->portal_group_tag, tag);
                            return 0;
                        }
                    }
                    else {
                        /* we now know the tag */
                        session->portal_group_tag = tag;
                    }
                    text = value_end;
                } 
                else if (iscsi_find_key_value("InitialR2T", text, end, &value, &value_end)) {
                    if (value && iscsi_strcmp(value, "Yes") == 0)
                        session->InitialR2T = 1;
                    else
                        session->InitialR2T = 0;
                    text = value_end;
                } 
                else if (iscsi_find_key_value("ImmediateData", text, end, &value, &value_end)) {
                    if (value && iscsi_strcmp(value, "Yes") == 0)
                        session->ImmediateData = 1;
                    else
                        session->ImmediateData = 0;
                    text = value_end;
                } 
                else if (iscsi_find_key_value("MaxRecvDataSegmentLength", text, end, &value, &value_end)) {
                    /* FIXME: no octal */
                    session->MaxXmitDataSegmentLength = iscsi_strtoul(value, NULL, 0);
                    text = value_end;
                } 
                else if (iscsi_find_key_value("FirstBurstLength", text, end, &value, &value_end)) {
                    /* FIXME: no octal */
                    session->FirstBurstLength = iscsi_strtoul(value, NULL, 0);
                    text = value_end;
                } 
                else if (iscsi_find_key_value("MaxBurstLength", text, end, &value, &value_end)) {
                    /* we don't really care, since it's a limit on the target's R2Ts, but record it anwyay */
                    /* FIXME: no octal, and draft16 says we MUST NOT send more than MaxBurstLength */
                    session->MaxBurstLength = iscsi_strtoul(value, NULL, 0);
                    text = value_end;
                } 
                else if (iscsi_find_key_value("HeaderDigest", text, end, &value, &value_end)) {
                    if (iscsi_strcmp(value, "None")) {
                        logmsg(AS_ERROR, "session %p can't accept %s\n", session, text);
                        return 0;
                    }
                    text = value_end;
                }
                else if (iscsi_find_key_value("DataDigest", text, end, &value, &value_end)) {
                    if (iscsi_strcmp(value, "None")) {
                        logmsg(AS_ERROR, "session %p can't accept %s\n", session, text);
                        return 0;
                    }
                    text = value_end;
                }
                else if (iscsi_find_key_value("DefaultTime2Wait", text, end, &value, &value_end)) {
                    session->DefaultTime2Wait = iscsi_strtoul(value, NULL, 0);
                    text = value_end;
                } 
                else if (iscsi_find_key_value("DefaultTime2Retain", text, end, &value, &value_end)) {
                    session->DefaultTime2Retain = iscsi_strtoul(value, NULL, 0);
                    text = value_end;
                } 
                else if (iscsi_find_key_value("OFMarker", text, end, &value, &value_end)) {
                    /* result function is AND, target must honor our No */
                    text = value_end;
                } 
                else if (iscsi_find_key_value("OFMarkInt", text, end, &value, &value_end)) {
                    /* we don't do markers, so we don't care */
                    text = value_end;
                } 
                else if (iscsi_find_key_value("IFMarker", text, end, &value, &value_end)) {
                    /* result function is AND, target must honor our No */
                    text = value_end;
                } 
                else if (iscsi_find_key_value("IFMarkInt", text, end, &value, &value_end)) {
                    /* we don't do markers, so we don't care */
                    text = value_end;
                } 
                else if (iscsi_find_key_value("DataPDUInOrder", text, end, &value, &value_end)) {
                    if (value && iscsi_strcmp(value, "Yes") == 0)
                        session->DataPDUInOrder = 1;
                    else
                        session->DataPDUInOrder = 0;
                    text = value_end;
                } 
                else if (iscsi_find_key_value("DataSequenceInOrder", text, end, &value, &value_end)) {
                    if (value && iscsi_strcmp(value, "Yes") == 0)
                        session->DataSequenceInOrder = 1;
                    else
                        session->DataSequenceInOrder = 0;
                    text = value_end;
                } 
                else if (iscsi_find_key_value("MaxOutstandingR2T", text, end, &value, &value_end)) {
                    if (iscsi_strcmp(value, "1")) {
                        logmsg(AS_ERROR, "session %p can't accept MaxOutstandingR2T %s\n", session, value);
                        return 0;
                    }
                    text = value_end;
                } 
                else if (iscsi_find_key_value("MaxConnections", text, end, &value, &value_end)) {
                    if (iscsi_strcmp(value, "1")) {
                        logmsg(AS_ERROR, "session %p can't accept MaxConnections %s\n", session, value);
                        return 0;
                    }
                    text = value_end;
                } 
                else if (iscsi_find_key_value("ErrorRecoveryLevel", text, end, &value, &value_end)) {
                    if (iscsi_strcmp(value, "0")) {
                        logmsg(AS_ERROR, "session %p can't accept ErrorRecovery %s\n", session, value);
                        return 0;
                    }
                    text = value_end;
                } 
                else if (iscsi_find_key_value("X-com.cisco.protocol", text, end, &value, &value_end)) {
                    if (iscsi_strcmp(value, "NotUnderstood") && iscsi_strcmp(value, "draft16")) {
                        /* if we didn't get a compatible protocol, fail */
                        logmsg(AS_ERROR, "can't accept protocol %s\n", value);
                        return 0;
                    }
                    text = value_end;
                }
                else if (iscsi_find_key_value("X-com.cisco.PingTimeout", text, end, &value, &value_end)) {
                    /* don't care */
                    text = value_end;
                } 
                else {
                    /* FIXME: we may want to ignore X- keys sent by
                     * the target, but that would require us to have
                     * another PDU buffer so that we can generate a
                     * response while we still know what keys we
                     * received, so that we can reply with a
                     * NotUnderstood response.  For now, reject logins
                     * with keys we don't understand.  Another option is
                     * to silently ignore them, and see if the target has
                     * a problem with that.  The danger there is we may
                     * get caught in an infinite loop where we send an empty
                     * PDU requesting a phase transition, and the target
                     * keeps sending an empty PDU denying a phase transition
                     * (because we haven't replied to it's key yet).
                     */
                    logmsg(AS_ERROR, "failed to recognize text %s\n", text);
                    return 0;
                }
                break;
            }
            default:
                return 0;
        }
    }

    if (session->current_phase == ISCSI_SECURITY_NEGOTIATION_PHASE) {
        switch (iscsiAuthClientRecvEnd(session->auth_client, null_callback, NULL, NULL)) {
            default:
            case iscsiAuthStatusNoError: /* treat this as an error, since we should get a different code */
            case iscsiAuthStatusError:
                logmsg(AS_ERROR, "session %p error authenticating with target %s\n", 
                       session, session->TargetName);
                return 0;

            case iscsiAuthStatusInProgress:
                /* this should only occur if we were authenticating the target,
                 * which we don't do yet, so treat this as an error.
                 */
                logmsg(AS_ERROR, "session %p error authenticating target %s\n", session, session->TargetName);
                return 0;
                
            case iscsiAuthStatusContinue:
                /* continue sending PDUs */
                break;

            case iscsiAuthStatusPass:
                logmsg(AS_INFO, "session %p authenticated by target %s\n", 
                       session, session->TargetName);
                break;
            case iscsiAuthStatusFail:
                logmsg(AS_ERROR, "session %p failed authentication with target %s\n", 
                       session, session->TargetName);
                return 0;
        }
    }

    /* record some of the PDU fields for later use */
    session->tsid = iscsi_ntohs(login_rsp_pdu->tsid);
    session->ExpCmdSn = iscsi_ntohl(login_rsp_pdu->expcmdsn);
    session->MaxCmdSn = iscsi_ntohl(login_rsp_pdu->maxcmdsn);
    if (login_rsp_pdu->status_class == STATUS_CLASS_SUCCESS)
        session->ExpStatSn = iscsi_ntohl(login_rsp_pdu->statsn) + 1;

    if (login_rsp_pdu->tbit) {
        /* advance to the next phase */
        session->partial_response = 0;
        session->current_phase = login_rsp_pdu->next;
    }
    else {
        /* we got a partial response, don't advance, more negotiation to do */
        session->partial_response = 1;
    }

    return 1;
}

int iscsi_make_text_pdu(iscsi_session_t *session, struct IscsiHdr *pdu, char *data, int max_data_length)
{
    struct IscsiTextHdr *text_pdu = (struct IscsiTextHdr *)pdu;

    /* initialize the PDU header */
    memset(text_pdu, 0, sizeof(*text_pdu));

    text_pdu->opcode = ISCSI_OP_TEXT_CMD;
    text_pdu->itt = iscsi_htonl(session->itt);
    text_pdu->ttt = RSVD_TASK_TAG;
    text_pdu->cmdsn = iscsi_htonl(session->CmdSn++); 
    text_pdu->expstatsn = iscsi_htonl(session->ExpStatSn);

    return 1;
}


int iscsi_make_login_pdu(iscsi_session_t *session, struct IscsiHdr *pdu, char *data, int max_data_length)
{
    int tbit = 0;
    char value[iscsiAuthStringMaxLength];
    struct IscsiLoginHdr *login_pdu = (struct IscsiLoginHdr *)pdu;

    /* initialize the PDU header */
    memset(login_pdu, 0, sizeof(*login_pdu));
    login_pdu->opcode = ISCSI_OP_LOGIN_CMD | ISCSI_OP_IMMEDIATE;
    login_pdu->cid = 0;
    memcpy(login_pdu->isid, session->isid, sizeof(session->isid));
    login_pdu->tsid = 0;
    login_pdu->cmdsn = iscsi_htonl(session->CmdSn); /* don't increment on immediate */
    
    login_pdu->min_version = ISCSI_DRAFT16_VERSION;
    login_pdu->max_version = ISCSI_DRAFT16_VERSION;

    /* we have to send 0 until full-feature phase */
    login_pdu->expstatsn = 0; /* iscsi_htonl(session->ExpStatSn); */

    /* the very first Login PDU has some additional requirements, 
     * and we need to decide what phase to start in.
     */
    if (session->current_phase == ISCSI_INITIAL_LOGIN_PHASE) {
        if (session->InitiatorName && session->InitiatorName[0]) {
            if (!iscsi_add_text(session, pdu, data, max_data_length, "InitiatorName", session->InitiatorName))
                return 0;
        }
        else {
            logmsg(AS_ERROR, "InitiatorName is required on the first Login PDU\n");
            return 0;
        }
        if (session->InitiatorAlias && session->InitiatorAlias[0]) {
            if (!iscsi_add_text(session, pdu, data, max_data_length, "InitiatorAlias", session->InitiatorAlias))
                return 0;
        }
        
        if (session->TargetName[0] != '\0') {
            if (!iscsi_add_text(session, pdu, data, max_data_length, "TargetName", session->TargetName))
                return 0;
        }

        if (!iscsi_add_text(session, pdu, data, max_data_length, "SessionType",
                                  (session->type == ISCSI_SESSION_TYPE_DISCOVERY) ? "Discovery" : "Normal"))
            return 0;
        
        if (session->auth_client) {
            /* we're prepared to do authentication */
            session->current_phase = session->next_phase = ISCSI_SECURITY_NEGOTIATION_PHASE;
        }
        else {
            /* can't do any authentication, skip that phase */
            session->current_phase = session->next_phase = ISCSI_OP_PARMS_NEGOTIATION_PHASE;
        }        
    }
                
    /* fill in text based on the phase */
    switch (session->current_phase) {
        case ISCSI_OP_PARMS_NEGOTIATION_PHASE: {
            if (!session->partial_response) {
                /* try to go to full feature phase */
                login_pdu->curr = session->current_phase = ISCSI_OP_PARMS_NEGOTIATION_PHASE;
                login_pdu->next = session->next_phase    = ISCSI_FULL_FEATURE_PHASE;
                login_pdu->tbit = 1;

                /* discovery sessions don't need most op params, but we may end up in this phase 
                 * if the target insists on it, or if we're skipping security phase.
                 */
                if (session->type == ISCSI_SESSION_TYPE_DISCOVERY)
                    break;

                /* request the desired settings the first time we are in this phase */
                if (!iscsi_add_text(session, pdu, data, max_data_length, "InitialR2T", session->InitialR2T ? "Yes" : "No"))
                    return 0;

                if (!iscsi_add_text(session, pdu, data, max_data_length, "ImmediateData", session->ImmediateData ? "Yes" : "No"))
                    return 0;

                iscsi_sprintf(value, "%d", session->MaxRecvDataSegmentLength);
                if (!iscsi_add_text(session, pdu, data, max_data_length, "MaxRecvDataSegmentLength", value))
                    return 0;

                iscsi_sprintf(value, "%d", session->MaxBurstLength);
                if (!iscsi_add_text(session, pdu, data, max_data_length, "MaxBurstLength", value))
                    return 0;

                iscsi_sprintf(value, "%d", session->FirstBurstLength);
                if (!iscsi_add_text(session, pdu, data, max_data_length, "FirstBurstLength", value))
                    return 0;

                iscsi_sprintf(value, "%d", session->DefaultTime2Wait);
                if (!iscsi_add_text(session, pdu, data, max_data_length, "DefaultTime2Wait", value))
                    return 0;
                
                iscsi_sprintf(value, "%d", session->DefaultTime2Retain);
                if (!iscsi_add_text(session, pdu, data, max_data_length, "DefaultTime2Retain", value))
                    return 0;

                /* these we must have */
                if (!iscsi_add_text(session, pdu, data, max_data_length, "HeaderDigest", "None"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "DataDigest", "None"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "MaxOutstandingR2T", "1"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "MaxConnections", "1"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "ErrorRecoveryLevel", "0"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "IFMarker", "No"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "OFMarker", "No"))
                    return 0;

                /* FIXME: the caller may want different settings for these. */
                if (!iscsi_add_text(session, pdu, data, max_data_length, "DataPDUInOrder", "Yes"))
                    return 0;
                if (!iscsi_add_text(session, pdu, data, max_data_length, "DataSequenceInOrder", "Yes"))
                    return 0;

                /* vendor-specific protocol specification. list of protocol level strings in order of preference 
                 * allowable values are: draft<n> (e.g. draft8), rfc<n> (e.g. rfc666).
                 * For example: 
                 *      "X-com.cisco.protocol=draft16,draft8" requests draft 16, or 8 if 16 isn't supported.
                 *      "X-com.cisco.protocol=draft8,draft16" requests draft 8, or 16 if 8 isn't supported.
                 *
                 * Targets that understand this key SHOULD return the protocol level they selected as
                 * a response to this key, though the active_version may be sufficient to distinguish
                 * which protocol was chosen.
                 *
                 * Note: This probably won't work unless we start in op param phase, since the security 
                 * phase limits what keys we can send, and we'd need to have sent this on the first PDU
                 * of the login.  Keep sending it for informational use, and so that we can sanity
                 * check things later if the RFC and draft16 are using the same active version number.
                 */
                if (!iscsi_add_text(session, pdu, data, max_data_length, "X-com.cisco.protocol", "draft16"))
                    return 0;
            }
            else {
                /* FIXME: echo back the keys the target sent us, with the current values for those keys
                 * or NotUnderstood, and request the next phase.
                 * FIXME: make this code handle vendor-defined keys sent by the target.
                 * We can't handle them now because we don't have anyplace to save the key until the 
                 * response is generated, which is what we'd need to do to send key=NotUnderstood.
                 */

                /* try to go to full feature phase */
                login_pdu->curr = session->current_phase;
                login_pdu->next = ISCSI_FULL_FEATURE_PHASE;
                login_pdu->tbit = 1;
            }
            break;
        }
        case ISCSI_SECURITY_NEGOTIATION_PHASE: {
            int keytype = iscsiAuthKeyTypeNone;
            int rc = iscsiAuthClientSendTransitBit(session->auth_client, &tbit);

            /* see if we're ready for a phase change */
            if (rc == iscsiAuthStatusNoError) {
                login_pdu->tbit = tbit;
                if (tbit)
                    /* for discovery sessions, try to go right to full feature phase, not op param phase */
                    if (session->type == ISCSI_SESSION_TYPE_DISCOVERY)
                        login_pdu->next = session->next_phase = ISCSI_FULL_FEATURE_PHASE;
                    else
                        login_pdu->next = session->next_phase = ISCSI_OP_PARMS_NEGOTIATION_PHASE;
                else
                    session->next_phase = ISCSI_SECURITY_NEGOTIATION_PHASE;
            }
            else
                return 0;
            
            /* enumerate all the keys the auth code might want to send */
            while (iscsiAuthClientGetNextKeyType(&keytype) == iscsiAuthStatusNoError) {
                int present = 0;
                char *key = (char *)iscsiAuthClientGetKeyName(keytype);
                
                /* add the key/value pairs the auth code wants to send */
                rc = iscsiAuthClientSendKeyValue(session->auth_client, keytype, &present, value);
                if ((rc == iscsiAuthStatusNoError) && present) {
                    if (!iscsi_add_text(session, pdu, data, max_data_length, key, value))
                        return 0;
                }
            }

            break;
        }
        case ISCSI_FULL_FEATURE_PHASE:
            logmsg(AS_ERROR, "can't send login PDUs in full feature phase\n");
            return 0;
        default:
            logmsg(AS_ERROR, "can't send login PDUs in unknown phase %d\n", session->current_phase);
            return 0;
    }
    
    return 1;
}


/* attempt to login to the target.
 * The caller must check the status class to determine if the login succeeded.
 * A return of 1 does not mean the login succeeded, it just means this function
 * worked, and the status class is valid info.  This allows the caller to decide
 * whether or not to retry logins, so that we don't have any policy logic here.
 */
int iscsi_login(iscsi_session_t *session, char *buffer, size_t bufsize, uint8_t *status_class, uint8_t *status_detail)
{
    int received_pdu = 0;
    int ret = 0;
    
    /* prepare the session */
    session->CmdSn = 1;
    session->ExpCmdSn = 1;
    session->MaxCmdSn = 1;
    session->ExpStatSn = 0;

    session->current_phase = ISCSI_INITIAL_LOGIN_PHASE;
    session->partial_response = 0;
    
    if (session->auth_client) {
        /* prepare for authentication */
        if (iscsiAuthClientInit(session->auth_client, iscsiAuthNodeTypeInitiator) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "couldn't init auth_client %p for session %p\n", session->auth_client, session);
            return 0;
        }

        if (iscsiAuthClientSetVersion(session->auth_client, iscsiAuthVersionRfc) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "couldn't set auth version RFC for session %p\n", session);
            goto done;
        }

        if (iscsiAuthClientSetUsername(session->auth_client, session->username) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "couldn't set username for session %p\n", session);
            goto done;
        }

        if (iscsiAuthClientSetPassword(session->auth_client, session->password, session->password_length) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "couldn't set password for session %p\n", session);
            goto done;
        }

        /* FIXME: disable the minimum size check for now */
        if (iscsiAuthClientSetIpSec(session->auth_client, 1) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "couldn't set password for session %p\n", session);
            goto done;
        }
        
        if (iscsiAuthClientSetAuthRemote(session->auth_client, 0) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "couldn't set auth remote for session %p\n", session);
            goto done;
        }
    }

    /* exchange PDUs until the login phase is complete, or an error occurs */
    do {
        struct IscsiHdr pdu;
        struct IscsiLoginRspHdr *login_rsp_pdu = (struct IscsiLoginRspHdr *)&pdu;
        char *data;
        int max_data_length;
        int timeout = 0;

        memset(buffer, 0, bufsize);

        data = buffer;
        max_data_length = bufsize;

        /* pick the appropriate timeout. If we know the target has
         * responded before, and we're in the security phase, we use a
         * longer timeout, since the authentication alogorithms can
         * take a while, especially if the target has to go talk to a
         * tacacs or RADIUS server (which may or may not be
         * responding).
         */
        if (received_pdu && (session->current_phase == ISCSI_SECURITY_NEGOTIATION_PHASE))
            timeout = session->auth_timeout;
        else
            timeout = session->login_timeout;

        /* fill in the PDU header and text data based on the login phase that we're in */
        if (!iscsi_make_login_pdu(session, &pdu, data, max_data_length))
            goto done;

        /* send a PDU to the target */
        if (!iscsi_send_pdu(session, &pdu, data, timeout))
            goto done;
        
        /* read the target's response into the same buffer */
        memset(buffer, 0, bufsize);
        if (!iscsi_recv_pdu(session, &pdu, sizeof(pdu), data, max_data_length, timeout))
            goto done;

        received_pdu = 1;

        /* check the PDU response type */
        if (pdu.opcode == (ISCSI_OP_LOGIN_RSP | 0xC0)) {
            /* it's a draft 8 login response, which we can't deal with */
            logmsg(AS_ERROR, "received iSCSI draft 8 login response opcode 0x%x, expected draft 16 login response 0x%2x\n",
                   pdu.opcode, ISCSI_OP_LOGIN_RSP);
            logmsg(AS_ERROR, "please make sure that you have installed the correct driver version.\n");
            goto done;
        }
        else if (pdu.opcode != ISCSI_OP_LOGIN_RSP) {
            logmsg(AS_ERROR, "received opcode 0x%2x during login, expected login response 0x%2x\n",
                   pdu.opcode, ISCSI_OP_LOGIN_RSP);
            goto done;
        }

        /* give the caller the status class and detail from the last login response PDU received */
        if (status_class)
            *status_class = login_rsp_pdu->status_class;
        if (status_detail)
            *status_detail = login_rsp_pdu->status_detail;

        if (login_rsp_pdu->status_class != STATUS_CLASS_SUCCESS) {
            /* this code didn't fail, but the login did.  
             * The caller must check the class and detail.
             * We process the login response anyway, so that
             * redirects can reset the session's IP address.
             */
            ret = 1; 
            iscsi_process_login_response(session, login_rsp_pdu, data, max_data_length);
            goto done;
        }
        else if (!iscsi_process_login_response(session, login_rsp_pdu, data, max_data_length))
            goto done;

    } while (session->current_phase != ISCSI_FULL_FEATURE_PHASE);

    ret = 1;

 done:
    if (session->auth_client) {
        if (iscsiAuthClientFinish(session->auth_client) != iscsiAuthStatusNoError) {
            logmsg(AS_ERROR, "error finishing authclient\n");
            return 0;
        }
    }

    return ret;
}
